#include <stdio.h>

#undef malloc
#undef free
#undef calloc
#undef realloc
#include <malloc.h>

extern char	*getenv (const char *);
extern FILE	*popen (const char *, const char *);
extern int	pclose (FILE *);

extern void	bt (FILE *);
extern int	btinit ();

static void	*(*old_malloc_hook) (size_t);
static void	(*old_free_hook) (void *);
static void	*(*old_realloc_hook) (void *, size_t);

static FILE	*mprfi;		/* filter */
static char	mprbuf[BUFSIZ];	/* its buffer */

static void *
malloc_hook (size_t sz)
{
	void *ptr;

	__malloc_hook = old_malloc_hook;
	ptr = malloc (sz);
	old_malloc_hook = __malloc_hook;
	__malloc_hook = malloc_hook;

	fprintf (mprfi, "m:");
	bt (mprfi);
	fprintf (mprfi, "%d:%ld\n", sz, (long) ptr);

	return ptr;
}

static void
free_hook (void *ptr)
{
	fprintf (mprfi, "f:");
	bt (mprfi);
	fprintf (mprfi, "%ld\n", (long) ptr);

	__free_hook = old_free_hook;
	free (ptr);
	old_free_hook = __free_hook;
	__free_hook = free_hook;
}

static void *
realloc_hook (void *ptr, size_t sz)
{
	void *ptr2;

	__realloc_hook = old_realloc_hook;
	__malloc_hook = old_malloc_hook;
	__free_hook = old_free_hook;
	ptr2 = realloc (ptr, sz);
	old_free_hook = __free_hook;
	old_malloc_hook = __malloc_hook;
	old_realloc_hook = __realloc_hook;
	__free_hook = free_hook;
	__malloc_hook = malloc_hook;
	__realloc_hook = realloc_hook;

	fprintf (mprfi, "r:");
	bt (mprfi);
	fprintf (mprfi, "%d:%ld:%ld\n", sz, (long) ptr, (long) ptr2);

	return ptr2;
}

int
mpr ()
{
	char *filter = getenv ("MPRFI");

	if (!filter || !*filter) /* disabled */
		return 0;

	if (btinit () < 0)
		return -1;

	if (!(mprfi = popen (filter, "w")))
		return -2;

	setbuf (mprfi, mprbuf); /* mprfi must avoid malloc */

	old_malloc_hook = __malloc_hook;
	old_free_hook = __free_hook;
	old_realloc_hook = __realloc_hook;
	__malloc_hook = malloc_hook;
	__free_hook = free_hook;
	__realloc_hook = realloc_hook;

	return 0;
}
